@@ -21,10 +21,13 @@ import os
 from . import utils
 import numpy as np
 import terminaltables
-
+import cryolo.preprocessing as preprocessing
 from . import config_tools
 from . import imagereader
 from gooey import Gooey, GooeyParser
+from cryolo import CoordsIO
+from typing import List
+from cryolo.utils import BoundBox, Dict
 
 
 """
@@ -62,7 +65,7 @@ def create_parser(parser):
     required_group = parser.add_argument_group(
         "Required arguments",
         "These options are required to run crYOLO evaluation. In most cases you want to use a runfile to evaluate"
-        "your model. In case you want to specify the ground truth data, leave the runfile field empty and go to the optional"
+        " your model. In case you want to specify the ground truth data, leave the runfile field empty and go to the optional"
         "arguments.",
     )
     required_group.add_argument(
@@ -110,7 +113,7 @@ def create_parser(parser):
     required_group.add_argument(
         "-o",
         "--output",
-        help="Select path where the results are written as .html ",
+        help="Select path to file where the results are written as .html ",
         default="result_evaluation.html",
         widget="FileSaver",
         gooey_options={
@@ -506,96 +509,12 @@ def main(args=None):
     print("Wrote evaluation results to", path_output)
 
 
-def prepare_datasets(
-    path_config, path_weights, path_boxfiles, path_runfile, path_images
-):
-    datasets = {}
 
-    if path_runfile:
-        # Read run file
-        with open(path_runfile) as runfile_buffer:
-            runfile = json.load(runfile_buffer)
-        images = runfile["run"]["valid_images"]
-        annot = runfile["run"]["valid_annot"]
-
-        for i in range(len(images)):
-            imgpath = images[i]
-            boxpath = annot[i]
-
-            if os.stat(boxpath).st_size == 0:
-                continue
-            last_folder = os.path.basename(os.path.dirname(imgpath))
-            if last_folder in datasets:
-                box_files = datasets[last_folder]["boxes"]
-                img_files = datasets[last_folder]["images"]
-            else:
-                box_files = []
-                img_files = []
-                datasets[last_folder] = {}
-                datasets[last_folder]["boxes"] = box_files
-                datasets[last_folder]["images"] = img_files
-            box_files.append(boxpath)
-            img_files.append(imgpath)
-    else:
-        # Scan all box filenames
-        for root, directories, filenames in os.walk(path_boxfiles):
-            if len(filenames) == 0:
-                continue
-            last_folder = os.path.basename(root)
-            if last_folder.startswith(("empty", ".")):
-                # Ignore empty grids
-                continue
-
-            box_files = []
-            for filename in filenames:
-                if filename.endswith(("box")) and not filename.startswith("."):
-                    box_files.append(os.path.join(root, filename))
-
-            if len(box_files) >= 1:
-                if last_folder in datasets:
-                    datasets[last_folder]["boxes"].append(box_files)
-                else:
-                    dataset = {"boxes": box_files}
-                    datasets[last_folder] = dataset
-
-        # Find all images
-        img_files = []
-        for root, directories, filenames in os.walk(path_images):
-            last_folder = os.path.basename(root)
-            if last_folder.startswith("empty"):
-                # Ignore empty grids
-                continue
-
-            for filename in filenames:
-                if filename.endswith(
-                    ("jpg", "png", "tiff", "tif", "mrc")
-                ) and not filename.startswith("."):
-                    img_files.append(os.path.join(root, filename))
-
-        # Find to boxifles corresponding images
-        for foldername, dataset in datasets.items():
-            boxfiles = dataset["boxes"]
-            imgfiles = []
-            for boxfile in boxfiles:
-                filename = os.path.splitext(os.path.basename(boxfile))[0]
-                lastfolder = os.path.basename(os.path.dirname(boxfile))
-                search_path = os.path.join(lastfolder, filename)
-                corresponding_img_path = difflib.get_close_matches(
-                    search_path, img_files, n=1, cutoff=0
-                )[0]
-                print("Assign", boxfile, "to", corresponding_img_path)
-                imgfiles.append(corresponding_img_path)
-            dataset["images"] = imgfiles
-
-    # Read config
-    with open(path_config) as config_buffer:
-        config = json.load(config_buffer)
-
-    # Configure model
+def get_model(config, path_weights, num_gpus):
     ###############################
     #   Make the model
     ###############################
-    num_gpus = 1
+    config["model"]["input_size"] = config_tools.get_adjusted_input_size(config)
     backend_weights = None
     anchors = utils.get_anchors(config, path_weights=path_weights)
     num_free_layers = utils.get_num_fine_tune_layers(path_weights)
@@ -635,9 +554,62 @@ def prepare_datasets(
 
         parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus)
         yolo.model = parallel_model
+    return yolo
 
+def prepare_datasets(
+    path_config, path_weights, path_boxfiles, path_runfile, path_images
+):
+    datasets = {}
+    pairs = []
+    if path_runfile:
+        # Read run file
+        with open(path_runfile) as runfile_buffer:
+            runfile = json.load(runfile_buffer)
+        images = runfile["run"]["valid_images"]
+        annot = runfile["run"]["valid_annot"]
+
+        for i in range(len(images)):
+            imgpath = images[i]
+            boxpath = annot[i]["path"]
+            pairs.append((imgpath,boxpath))
+
+    else:
+        pairs = preprocessing.find_image_annotation_pairs_by_dir(path_boxfiles,path_images)
+
+    for pair in pairs:
+        imgpath = pair[0]
+        boxpath = pair[1]
+        print(pair)
+        if os.stat(boxpath).st_size == 0:
+            continue
+        last_folder = os.path.basename(os.path.dirname(imgpath))
+        width, height = imagereader.read_width_height(imgpath)
+        key = "{}_{}_{}".format(last_folder, width, height)
+        if key in datasets:
+            box_files = datasets[key]["boxes"]
+            img_files = datasets[key]["images"]
+        else:
+            box_files = []
+            img_files = []
+            datasets[key] = {}
+            datasets[key]["boxes"] = box_files
+            datasets[key]["images"] = img_files
+        box_files.append(boxpath)
+        img_files.append(imgpath)
+
+    # Read config
+    with open(path_config) as config_buffer:
+        config = json.load(config_buffer)
+
+    # Configure model
+
+
+    import copy
     # Get picks
     for foldername, dataset in datasets.items():
+
+        #yolo = get_model(config=config, path_weights=path_weights, num_gpus=1)
+
         # Get image dimensions of current dataset
         width, height = imagereader.read_width_height(dataset["images"][0])
 
@@ -645,7 +617,7 @@ def prepare_datasets(
         one_box_file = read_box(dataset["boxes"][0], height)
         box_size = one_box_file[0]["w"]
         # Adapt the anchor size in config file.
-        config["model"]["particle_diameter"] = box_size
+        #config["model"]["particle_diameter"] = box_size
         # if is_single_anchor:
         #     anchor_width = 1.0 * box_size / cell_w
         #     anchor_height = 1.0 * box_size / cell_h
@@ -664,27 +636,33 @@ def prepare_datasets(
         # Predict
         from cryolo import predict
 
-        predict_results = predict.do_prediction(
+        #  Get normalization option
+        normalization_string = "STANDARD"
+        if "norm" in config["model"]:
+            normalization_string = config["model"]["norm"]
+        print("IMG:",dataset["images"])
+        predict_results, _ = predict.do_prediction(
             config_path=path_config,
             weights_path=path_weights,
             num_patches=num_patches,
             input_path=dataset["images"],
             obj_threshold=0.0,
-            config_pre=config,
+            config_pre=copy.deepcopy(config),
             overlap=overlap_patches,
-            yolo=yolo,
+            yolo=None,
             min_size=None,
             max_size=None,
+            normalization=normalization_string
         )
-
         # Rescale to image dimensions
         for boxes_in_micrograph in predict_results:
             image_width = boxes_in_micrograph["img_width"]
             image_height = boxes_in_micrograph["img_height"]
-            for box in boxes_in_micrograph["boxes"]:
-                predict.rescale(box, image_height, image_width, box_size)
-                box.x = box.x + box_size / 2
-                box.y = box.y + box_size / 2
+            for slice_number, boxes_in_slice in enumerate(boxes_in_micrograph["boxes"]):
+                for box in boxes_in_slice:
+                    predict.rescale(box, image_height, image_width, box_size)
+                    box.x = box.x + box_size / 2
+                    box.y = box.y + box_size / 2
 
         dataset["picks"] = predict_results
 
@@ -804,7 +782,7 @@ def get_precision_recall_f1(dataset, confidence_threshold):
     iou_tresh = 0.5
     picks = dataset["picks"]
 
-    boxes = dataset["boxes"]
+    boxes = dataset["boxes"] # boxfile paths
     images = dataset["images"]
     sum_picked = 0
     images_with_particles = 0
@@ -814,7 +792,13 @@ def get_precision_recall_f1(dataset, confidence_threshold):
         gd_box_remove = []
         picked_boxes_index_remove = []
 
-        picked_boxes = boxes_in_mic["boxes"]
+        picked_boxes = boxes_in_mic["boxes"] # picked boxes... what a confusion. stupid me.
+        if len(picked_boxes) == 1:
+            picked_boxes = boxes_in_mic["boxes"][0]
+        else:
+            print("Not implemented for tomography yet. Exit.")
+            import sys
+            sys.exit()
         img_path = boxes_in_mic["img_path"]
         img_name_without_suffix = os.path.splitext(os.path.basename(img_path))[0]
         # ground_truth_boxes_path = difflib.get_close_matches(img_name_without_suffix, boxes, n=1, cutoff=0)[0]
@@ -1015,22 +999,23 @@ def calc_ious(ground_truth_box, picked_boxes):
     return [bbox_iou(ground_truth_box, p_box) for p_box in picked_boxes]
 
 
-def read_box(path, image_height):
+def read_box(path, image_height) -> List[Dict]:
     boxes = []
     # with open(path, 'r') as boxfile:
-    boxreader = np.atleast_2d(
-        np.genfromtxt(path)
-    )  # csv.reader(boxfile, delimiter='\t',
-    #          quotechar='|', quoting=csv.QUOTE_NONE)
+    if path.endswith(".cbox"):
+        all_boxes = CoordsIO.read_cbox_boxfile(path)
+    elif path.endswith(".box"):
+        all_boxes = CoordsIO.read_eman1_boxfile(path)
 
     try:
-        for box in boxreader:
+        for box in all_boxes:
             box_d = {}
-            width = int(float(box[2]))
-            height = int(float(box[3]))
 
-            box_d["x"] = int(float(box[0])) + width / 2
-            box_d["y"] = int(float(box[1])) + height / 2
+            width = int(float(box.w))
+            height = int(float(box.h))
+
+            box_d["x"] = int(float(box.x)) + width / 2
+            box_d["y"] = int(float(box.y)) + height / 2
             box_d["w"] = width
             box_d["h"] = height
             box_d["filename"] = path
